//===-- driver/ldc-version.c - ----------------------------------*- C++ -*-===//
//
//                         LDC – the LLVM D compiler
//
// This file is distributed under the BSD-style LDC license. See the LICENSE
// file for details.
//
//===----------------------------------------------------------------------===//

#include "driver/ldc-version.h"

namespace ldc {

const char * const ldc_version = "1.42.0";
const char * const dmd_version = "v2.112.1";
const char * const llvm_version = "21.1.8";
const char * const llvm_version_base = "21.1.8";
const char * const built_with_Dcompiler_version = "LDC - the LLVM D compiler (1.41.0)";

}
